/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2014                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                              HC_PROCO4                                 **
**                      Wersja z dnia : 2018.09.01                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/


#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include "proc4eeprom.h"


#define false                          0
#define true                           1

#define LEDPort                        PORTA
#define LEDPortDirection               DDRA

#define StatusLedPosInPort             3
#define RS232RxTxLedPosInPort          4
#define RS485RxTxLedPosInPort          5
#define EEWriteLedPosInPort            6
#define ConnectionLedPosInPort         7

#define TimerCounterDelayLimit         3000  // 2048
#define SubDevTimerCounterLimit        1024  // 2048
#define QueueTimerCounterLimit         512   // 1024
#define StandardLEDPulseDelay          244   // 832

#define RxTxPort                       PORTD
#define RxTxDirectionPort              DDRD
#define RxTxDirPin                     4              // PORTD . 4

#define RTSPin                         6
#define CTSPin                         5
#define ModemControlPort               PORTD
#define ModemStatusPort                PIND
#define ModemDirPort                   DDRD

#define null                           0
#define cr                             0x0D
#define Space                          0x20
#define PartSeparator                  ':'
#define ParamSeparator                 '.'

#define nop() __asm__ __volatile__ ("nop")

// USART BAUD Rate Generator fosc=7.3728 MHz
#define Speed9600bps                   47
#define Speed19200bps                  23
#define Speed38400bps                  11
#define Speed115200bps                 3

#define IdleState                      0
#define BusyState                      1

#define TXD232BuffSize                 64
#define RXD232BuffSize                 16

typedef struct { volatile uint8_t  TXDCycFr ;
                 volatile uint8_t  TXDCycTo ;
                 volatile uint8_t  TXDState ;
                 volatile uint8_t  RXDCycFr ;
                 volatile uint8_t  RXDCycTo ;
                 volatile uint16_t LEDDelay ;
                 uint8_t           TXDBuffer [ TXD232BuffSize ] ;
                 uint8_t           RXDBuffer [ RXD232BuffSize ] ;
               } USART232RecT ;

#define TXD485BuffSize                 200
#define RXD485BuffSize                 16

typedef struct { volatile uint8_t  TXDCycFr ;
                 volatile uint8_t  TXDCycTo ;
                 volatile uint8_t  TXDState ;
                 volatile uint8_t  RXDCycFr ;
                 volatile uint8_t  RXDCycTo ;
                 volatile uint16_t LEDDelay ;
                 uint8_t           TXDBuffer [ TXD485BuffSize ] ;
                 uint8_t           RXDBuffer [ RXD485BuffSize ] ;
               } USART485RecT ;


#define CommandBufferSize              16
typedef uint8_t CommandBufferType [ CommandBufferSize + 1 ] ;
#define ParamBufferSize                32
typedef uint8_t ParamBufferType [ ParamBufferSize + 1 ] ;

struct _SerialInstanceStruct ;

typedef struct _SerialInstanceStruct * PtrSerialInstanceRecType ;

typedef void ( * ServiceProcType ) ( void ) ;

typedef void ( * SendSerialProcT ) ( uint8_t /* SerialCh */ ) ;

typedef uint8_t ( * SerialDataPresentProcT ) ( void ) ;

typedef uint8_t ( * GetSerialProcT ) ( void ) ;

typedef void ( * SerialOperationProcT ) ( PtrSerialInstanceRecType /* SerialInstance */ ) ;

typedef struct _SerialInstanceStruct {
                 SendSerialProcT           EchoSerial ;
                 SendSerialProcT           SendSerial ;
                 SerialDataPresentProcT    SerialDataPresent ;
                 GetSerialProcT            GetSerial ;
                 SerialOperationProcT      SerialService ;
                 ServiceProcType           ExecuteCommand ;
                 CommandBufferType         CommandText ;
                 ParamBufferType           ParamText ;
                 DeviceIDType              DestinationDeviceName ;
                 DeviceIDType              SourceDeviceName ;
                 uint8_t                   CommandTextIndex ;
                 uint8_t                   ParamTextIndex ;
                 uint8_t                   DestDeviceNameIndex ;
                 uint8_t                   SrcDeviceNameIndex ;
               } SerialInstanceRecType ;

typedef struct {
                 uint8_t                   ValidDescr ;
                 uint8_t                   CurrentState ;
                 DeviceIDType              SubDeviceID ;
               } ChannalsDataRecT ;

#define BuffQueueSize                  2
#define SendQueueSize                  2

#define TimeEvent                      0x01
#define SubDeviceEvent                 0x02
#define QueueEvent                     0x04

static USART232RecT USART232 ;
static USART485RecT USART485 ;
volatile static uint8_t Events ;
static SerialInstanceRecType SerialRS232Instance ;
static SerialInstanceRecType SerialRS485Instance ;
static uint8_t ConnectContext ;
static uint8_t CurrentTimeStatusState ;
static uint16_t TimerCounter ;
static uint16_t SubDevTimerCounter ;
static uint8_t SubDeviceIndex ;
static ChannalsDataRecT ChannalsData [ ChannelsArraySize ] ;
static uint16_t QueueTimerCounter ;


static uint8_t DeviceName [ ] PROGMEM               = "HC-PROCO4 - procesor komunikacyjny\r\n" ;
static uint8_t HardwareSoftwareVersion [ ] PROGMEM  = "HVER*1_0*IX-2011.SVER*1_1*VI-2018\r" ;
static uint8_t EmptyData [ ] PROGMEM                = "<EMPTY>\r" ;
static uint8_t OK [ ] PROGMEM                       = "<OK>\r" ;
static uint8_t Fail [ ] PROGMEM                     = "<FAIL>\r" ;
static uint8_t GeneralPoolDeviceName [ ] PROGMEM    = "GENERAL" ;

static uint8_t HelloCommand [ ] PROGMEM             = "HELLO" ;
static uint8_t ConnectCommand [ ] PROGMEM           = "CONNECT" ;
static uint8_t DisconnectCommand [ ] PROGMEM        = "DISCONNECT" ;
static uint8_t SetDeviceNameCommand [ ] PROGMEM     = "SETMID" ;
static uint8_t AddSuperiorChnCommand [ ] PROGMEM    = "ADDSUP" ;
static uint8_t DelSuperiorChnCommand [ ] PROGMEM    = "DELSUP" ;
static uint8_t ListSuperiorChnCommand [ ] PROGMEM   = "LISTSUP" ;
static uint8_t AddSlaveIDCommand [ ] PROGMEM        = "ADDSL" ;
static uint8_t DelSlaveIDCommand [ ] PROGMEM        = "DELSL" ;
static uint8_t ListSlaveIDCommand [ ] PROGMEM       = "LISTSL" ;
static uint8_t GetVersionCommand [ ] PROGMEM        = "GETVER" ;
static uint8_t GetIDCommand [ ] PROGMEM             = "GETID" ;
static uint8_t ReinitCommand [ ] PROGMEM            = "REINIT" ;
static uint8_t ResetEECommand [ ] PROGMEM           = "RESEE" ;
static uint8_t SaveEECommand [ ] PROGMEM            = "SAVEEE" ;
static uint8_t NewPasswordCommand [ ] PROGMEM       = "NEWPW" ;
static uint8_t EchoOnCommand [ ] PROGMEM            = "ECHOON" ;
static uint8_t EchoOffCommand [ ] PROGMEM           = "ECHOOFF" ;

static uint8_t ReplyReqCommand [ ] PROGMEM          = "REPLY" ;

typedef void ( * CommandExecuteServiceProcT ) ( PtrSerialInstanceRecType SerialInstance ) ;
typedef struct {
                 CommandExecuteServiceProcT  CommandService ;
                 uint8_t *                   CommandName ;
               } CommandNameAssociationRecT ;

static void HelloCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void ConnectCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void DisconnectCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void SetDeviceNameCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void AddSuperiorChnCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void DelSuperiorChnCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void ListSuperiorChnCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void AddSlaveIDCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void DelSlaveIDCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void ListSlaveIDCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void GetVersionCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void GetIDCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void ReinitCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void ResetEECommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void SaveEECommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void NewPasswordCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void EchoOnCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;
static void EchoOffCommandExecute ( PtrSerialInstanceRecType SerialInstance ) ;

#define SvCommandNameAssociationTabeSize             18

static CommandNameAssociationRecT SvCommandNameAssociationTabe [ SvCommandNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { HelloCommandExecute                 , HelloCommand } ,
/* 01 */ { ConnectCommandExecute               , ConnectCommand } ,
/* 02 */ { DisconnectCommandExecute            , DisconnectCommand } ,
/* 03 */ { SetDeviceNameCommandExecute         , SetDeviceNameCommand } ,
/* 04 */ { GetVersionCommandExecute            , GetVersionCommand } ,
/* 05 */ { NewPasswordCommandExecute           , NewPasswordCommand } ,
/* 06 */ { GetIDCommandExecute                 , GetIDCommand } ,
/* 07 */ { AddSuperiorChnCommandExecute        , AddSuperiorChnCommand } ,
/* 08 */ { DelSuperiorChnCommandExecute        , DelSuperiorChnCommand } ,
/* 09 */ { ListSuperiorChnCommandExecute       , ListSuperiorChnCommand } ,
/* 10 */ { AddSlaveIDCommandExecute            , AddSlaveIDCommand } ,
/* 11 */ { DelSlaveIDCommandExecute            , DelSlaveIDCommand } ,
/* 12 */ { ListSlaveIDCommandExecute           , ListSlaveIDCommand } ,
/* 13 */ { ReinitCommandExecute                , ReinitCommand } ,
/* 14 */ { ResetEECommandExecute               , ResetEECommand } ,
/* 15 */ { SaveEECommandExecute                , SaveEECommand } ,
/* 16 */ { EchoOnCommandExecute                , EchoOnCommand } ,
/* 17 */ { EchoOffCommandExecute               , EchoOffCommand } ,
} ;


typedef enum { StatusLED ,
               RS232RxTxLED ,
               RS485RxTxLED ,
               ConnectLED ,
               EEWriteLED
             } LEDSinalT ;


static void LEDOn ( LEDSinalT LEDSignal )
{
  /*----------------------------------------------------------------*/
  switch ( LEDSignal )
  {
    case StatusLED :
      LEDPort &= ~ ( 1 << StatusLedPosInPort ) ;
      break ;
    case RS232RxTxLED :
      LEDPort &= ~ ( 1 << RS232RxTxLedPosInPort ) ;
      break ;
    case RS485RxTxLED :
      LEDPort &= ~ ( 1 << RS485RxTxLedPosInPort ) ;
      break ;
    case ConnectLED :
      LEDPort &= ~ ( 1 << ConnectionLedPosInPort ) ;
      break ;
    case EEWriteLED :
      LEDPort &= ~ ( 1 << EEWriteLedPosInPort ) ;
      break ;
  } /* switch */ ;
} /* LEDOn */



static void LEDOff ( LEDSinalT LEDSignal )
{
  /*----------------------------------------------------------------*/
  switch ( LEDSignal )
  {
    case StatusLED :
      LEDPort |= ( 1 << StatusLedPosInPort ) ;
      break ;
    case RS232RxTxLED :
      LEDPort |= ( 1 << RS232RxTxLedPosInPort ) ;
      break ;
    case RS485RxTxLED :
      LEDPort |= ( 1 << RS485RxTxLedPosInPort ) ;
      break ;
    case ConnectLED :
      LEDPort |= ( 1 << ConnectionLedPosInPort ) ;
      break ;
    case EEWriteLED :
      LEDPort |= ( 1 << EEWriteLedPosInPort ) ;
      break ;
  } /* switch */ ;
} /* LEDOff */


static void StartUSART232RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( RS232RxTxLED ) ;
  USART232 . LEDDelay = StandardLEDPulseDelay ;
} /* StartUSART232RxTxLED */


static void StopUSART232RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOff ( RS232RxTxLED ) ;
} /* StopUSART232RxTxLED */


static void StartUSART485RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( RS485RxTxLED ) ;
  USART485 . LEDDelay = StandardLEDPulseDelay ;
} /* StartUSART485RxTxLED */


static void StopUSART485RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOff ( RS485RxTxLED ) ;
} /* StopUSART485RxTxLED */


ISR ( TIMER0_OVF_vect )
{
  /*----------------------------------------------------------------*/
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    TimerCounter = 0 ;
    Events |= TimeEvent ;
  } /* if */ ;
  if ( USART232 . LEDDelay )
  {
    USART232 . LEDDelay -- ;
    if ( ! USART232 . LEDDelay )
    {
      StopUSART232RxTxLED ( ) ;
    } /* if */ ;
  } /* if */ ;
  if ( USART485 . LEDDelay )
  {
    USART485 . LEDDelay -- ;
    if ( ! USART485 . LEDDelay )
    {
      StopUSART485RxTxLED ( ) ;
    } /* if */ ;
  } /* if */ ;
  SubDevTimerCounter ++ ;
  if ( SubDevTimerCounter > SubDevTimerCounterLimit )
  {
    SubDevTimerCounter = 0 ;
    Events |= SubDeviceEvent ;
  } /* if */ ;
  QueueTimerCounter ++ ;
  if ( QueueTimerCounter > QueueTimerCounterLimit )
  {
    QueueTimerCounter = 0 ;
    Events |= QueueEvent ;
  } /* if */ ;
} /* TIMER0_OVF_vect */


static void USART485ReceiveMode ( void )
{
  /*----------------------------------------------------------------*/
  RxTxPort &= ~ ( 1 << RxTxDirPin ) ;
} /* USART485ReceiveMode */


static void USART485TransmitMode ( void )
{
  /*----------------------------------------------------------------*/
  RxTxPort |= ( 1 << RxTxDirPin ) ;
} /* USART485TransmitMode */


ISR ( USART0_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR0 ;
  USART232 . RXDBuffer [ USART232 . RXDCycTo ] = RecCh ;
  USART232 . RXDCycTo ++ ;
  if ( USART232 . RXDCycTo >= RXD232BuffSize )
    USART232 . RXDCycTo = 0 ;
  StartUSART232RxTxLED ( ) ;
} /* USART0_RX_vect */


ISR ( USART0_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( USART232 . TXDCycFr == USART232 . TXDCycTo )
  {
    USART232 . TXDState = IdleState ;
  } /* if ... */
  else
  {
    UDR0 = USART232 . TXDBuffer [ USART232 . TXDCycFr ] ;
    USART232 . TXDCycFr ++ ;
    if ( USART232 . TXDCycFr >= TXD232BuffSize )
      USART232 . TXDCycFr = 0 ;
    StartUSART232RxTxLED ( ) ;
  } /* if ... else */ ;
} /* USART0_TX_vect */


ISR ( USART1_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR1 ;
  USART485 . RXDBuffer [ USART485 . RXDCycTo ] = RecCh ;
  USART485 . RXDCycTo ++ ;
  if ( USART485 . RXDCycTo >= RXD485BuffSize )
    USART485 . RXDCycTo = 0 ;
  StartUSART485RxTxLED ( ) ;
} /* USART1_RX_vect */


ISR ( USART1_TX_vect )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( USART485 . TXDCycFr == USART485 . TXDCycTo )
  {
    USART485 . TXDState = IdleState ;
    USART485ReceiveMode ( ) ;
  } /* if ... */
  else
  {
    Ch = USART485 . TXDBuffer [ USART485 . TXDCycFr ] ;
    UDR1 = Ch ;
    USART485 . TXDCycFr ++ ;
    if ( USART485 . TXDCycFr >= TXD485BuffSize )
      USART485 . TXDCycFr = 0 ;
    StartUSART485RxTxLED ( ) ;
  } /* if ... else */ ;
} /* USART1_TX_vect */


static void USART232SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  cli ( ) ;
  StartUSART232RxTxLED ( ) ;
  if ( USART232 . TXDState == IdleState )
  {
    USART232 . TXDState = BusyState ;
    UDR0 = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = USART232 . TXDCycTo + 1 ;
      if ( NextInx >= TXD232BuffSize )
        NextInx = 0 ;
      if ( NextInx != USART232 . TXDCycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    USART232 . TXDBuffer [ USART232 . TXDCycTo ] = SerialCh ;
    USART232 . TXDCycTo ++ ;
    if ( USART232 . TXDCycTo >= TXD232BuffSize )
      USART232 . TXDCycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* USART232SendSerial */


static uint8_t USART232DataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( USART232 . RXDCycFr == USART232 . RXDCycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* USART232DataPresent */


static uint8_t USART232GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = USART232 . RXDBuffer [ USART232 . RXDCycFr ] ;
  USART232 . RXDCycFr ++ ;
  if ( USART232 . RXDCycFr >= RXD232BuffSize )
    USART232 . RXDCycFr = 0 ;
  return ( Ch ) ;
} /* USART232GetSerial */


static void USART485SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  cli ( ) ;
  NextInx = USART485 . TXDCycTo + 1 ;
  if ( NextInx >= TXD485BuffSize )
    NextInx = 0 ;
  if ( NextInx != USART485 . TXDCycFr )
  {
    USART485 . TXDBuffer [ USART485 . TXDCycTo ] = SerialCh ;
    USART485 . TXDCycTo ++ ;
    if ( USART485 . TXDCycTo >= TXD485BuffSize )
      USART485 . TXDCycTo = 0 ;
  } /* if */ ;
  sei ( ) ;
} /* USART485SendSerial */


static uint8_t USART485DataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( USART485 . RXDCycFr == USART485 . RXDCycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* USART485DataPresent */


static uint8_t USART485GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = USART485 . RXDBuffer [ USART485 . RXDCycFr ] ;
  USART485 . RXDCycFr ++ ;
  if ( USART485 . RXDCycFr >= RXD485BuffSize )
    USART485 . RXDCycFr = 0 ;
  return ( Ch ) ;
} /* USART485DataPresent */


static void USART0HardwareInit ( uint16_t RS232SerialSpeed )
{
  /*----------------------------------------------------------------*/
  UBRR0H = ( uint8_t ) ( RS232SerialSpeed >> 8 ) ;
  UBRR0L = ( uint8_t ) ( RS232SerialSpeed & 0xFF ) ;
  UCSR0A = 0 ;
  UCSR0B = ( 1 << RXCIE0 ) | ( 1 << TXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) ;
  UCSR0C = ( 1 << UCSZ01 ) | ( 1 << UCSZ00 ) ;
} /* USART0HardwareInit */


static void USART1HardwareInit ( uint16_t RS485SerialSpeed )
{
  /*----------------------------------------------------------------*/
  UBRR1H = ( uint8_t ) ( RS485SerialSpeed >> 8 ) ;
  UBRR1L = ( uint8_t ) ( RS485SerialSpeed & 0xFF ) ;
  UCSR1A = 0 ;
  UCSR1B = ( 1 << RXCIE1 ) | ( 1 << TXCIE1 ) | ( 1 << RXEN1 ) | ( 1 << TXEN1 ) ;
  UCSR1C = ( 1 << UCSZ11 ) | ( 1 << UCSZ10 ) ;
} /* USART1HardwareInit */


static void USART0Reset ( void )
{
  /*----------------------------------------------------------------*/
  USART232 . RXDCycFr = 0 ;
  USART232 . RXDCycTo = 0 ;
  USART232 . TXDCycFr = 0 ;
  USART232 . TXDCycTo = 0 ;
  USART232 . TXDState = IdleState ;
  USART232 . LEDDelay = 0 ;
} /* USART0Reset */


static void USART1Reset ( void )
{
  /*----------------------------------------------------------------*/
  USART485 . RXDCycFr = 0 ;
  USART485 . RXDCycTo = 0 ;
  USART485 . TXDCycFr = 0 ;
  USART485 . TXDCycTo = 0 ;
  USART485 . TXDState = IdleState ;
  USART485 . LEDDelay = 0 ;
} /* USART1Reset */


static void HardwareInit ( uint16_t RS485SerialSpeed ,
                           uint16_t RS232SerialSpeed )
{
  /*----------------------------------------------------------------*/
  TIMSK0 = ( 1 << TOIE0 ) ;
  TCCR0A = 0 ;
  TCCR0B = ( 1 << CS01 ) ; // CS02=0 CS01=1 CS00=0 : F clk div 8
  TCNT0 = 0 ;
  USART0HardwareInit ( RS232SerialSpeed ) ;
  USART1HardwareInit ( RS485SerialSpeed ) ;
} /* HardwareInit */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << StatusLedPosInPort ) | ( 1 << RS232RxTxLedPosInPort ) |
         ( 1 << RS485RxTxLedPosInPort ) | ( 1 << EEWriteLedPosInPort ) |
         ( 1 << ConnectionLedPosInPort ) ;
  LEDPortDirection |= Bits ;
  LEDPort |= Bits ;
  Bits = ( 1 << RxTxDirPin ) ;
  RxTxDirectionPort |= Bits ;
  USART485TransmitMode ( ) ;
  ModemDirPort |= ( 1 << CTSPin ) ;
  ModemDirPort &= ~ ( ( 1 << RTSPin ) ) ;
  ModemControlPort |= ( 1 << CTSPin ) ;
} /* EnvirInit */


static void SendTxt ( PtrSerialInstanceRecType SerialInstance ,
                      uint8_t *                String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      SerialInstance -> SendSerial ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* SendTxt */


static void SendFlashTxt ( PtrSerialInstanceRecType SerialInstance ,
                           uint16_t                 String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      SerialInstance -> SendSerial ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* SendFlashTxt */


static void NoEcho ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
} /* NoEcho */


static void ClearBlock ( void *   Destination ,
                         uint16_t Size )
{
  uint16_t Loop ;
  uint8_t * Dst ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
    * Dst ++ = null ;
} /* ClearBlock */


static void StrCopy ( uint8_t * Destination ,
                      uint8_t * Source )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = * Source ;
  * Destination = Ch ;
  while ( Ch )
  {
    Ch = * Source ++ ;
    * Destination ++ = Ch ;
  } /* while */ ;
} /* StrCopy */


static void MemCopy ( void *  Destination ,
                      void *  Source ,
                      uint8_t Size )
{
  uint8_t Loop ;
  uint8_t * Dst ;
  uint8_t * Src ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  Src = ( uint8_t * ) Source ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
  {
    * Dst ++ = * Src ++ ;
  } /* for */ ;
} /* MemCopy */


static void CodePassword ( void )
{
  uint8_t Loop ;
  uint8_t Element ;
  uint8_t BitCorrection ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = EEBlock . Password [ Loop ] ^ 0xFF ;
    BitCorrection = Element & 0x01 ;
    Element = Element >> 1 ;
    if ( BitCorrection )
      Element |= 0x80 ;
    EEBlock . Password [ Loop ] = Element ;
  } /* for */ ;
} /* CodePassword */


static void DecodePassword ( void )
{
  uint8_t Loop ;
  uint16_t Element ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = ( uint16_t ) EEBlock . Password [ Loop ] ;
    Element = Element << 1 ;
    if ( Element >= 0x100 )
      Element ++ ;
    Element &= 0xFF ;
    Element ^= 0xFF ;
    EEBlock . Password [ Loop ] = ( uint8_t ) Element ;
  } /* for */ ;
} /* DecodePassword */


static void ReadEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  eeprom_read_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                      sizeof ( EEPROMBlockDataRecT ) ) ;
  DecodePassword ( ) ;
} /* ReadEEBlock */


static void WriteEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( EEWriteLED ) ;
  CodePassword ( ) ;
  wdt_disable ( ) ;
  eeprom_write_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                       sizeof ( EEPROMBlockDataRecT ) ) ;
  wdt_enable ( WDTO_1S ) ;
  LEDOff ( EEWriteLED ) ;
  DecodePassword ( ) ;
} /* WriteEEBlock */


static uint16_t CRCTable ( uint8_t Index )
{
  /*----------------------------------------------------------------*/
  switch ( Index )
  {
    case 0 :
      return 0x0000 ;
    case 1 :
      return 0xCC01 ;
    case 2 :
      return 0xD801 ;
    case 3 :
      return 0x1400 ;
    case 4 :
      return 0xF001 ;
    case 5 :
      return 0x3C00 ;
    case 6 :
      return 0x2800 ;
    case 7 :
      return 0xE401 ;
    case 8 :
      return 0xA001 ;
    case 9 :
      return 0x6C00 ;
    case 10 :
      return 0x7800 ;
    case 11 :
      return 0xB401 ;
    case 12 :
      return 0x5000 ;
    case 13 :
      return 0x9C01 ;
    case 14 :
      return 0x8801 ;
    case 15 :
      return 0x4400 ;
  } /* switch */ ;
  return 0 ;
} /* CRCTable */


static uint16_t ComputeCRC ( uint16_t  Length ,
                             uint8_t * Buffer )
{
  uint16_t Loop ;
  uint16_t ResultCRC ;
  uint16_t WorkCRC ;
  /*----------------------------------------------------------------*/
  ResultCRC = 0xFFFF ;
  for ( Loop = 0 ; Loop < Length ; Loop ++ )
  {
    WorkCRC = CRCTable ( ResultCRC & 0x0F ) ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable ( Buffer [ Loop ] & 0x0F ) ;
    WorkCRC = CRCTable ( ResultCRC & 0x0F ) ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable ( ( Buffer [ Loop ] >> 4 ) & 0x0F ) ;
  } /* for */ ;
  return ( ResultCRC ) ;
} /* ComputeCRC */


static void EEBlockOperation ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  ReadEEBlock ( ) ;
  if ( EEBlock . EEBlockMetric != StandardSygnature )
  {
    SetDefaultEEBlock ( ) ;
    CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                            sizeof ( uint16_t ) ,
                            ( uint8_t * ) ( & EEBlock ) ) ;
    EEBlock . CRCValue = CRCValue ;
    WriteEEBlock ( ) ;
    return ;
  } /* if */ ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  if ( CRCValue != 0 )
  {
    SetDefaultEEBlock ( ) ;
    CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                            sizeof ( uint16_t ) ,
                            ( uint8_t * ) ( & EEBlock ) ) ;
    EEBlock . CRCValue = CRCValue ;
    WriteEEBlock ( ) ;
  } /* if */ ;
} /* EEBlockOperation */


static void InitSingleChannel ( ChannalsDataRecT * ChData ,
                                uint8_t *          DeviceID )
{
  /*----------------------------------------------------------------*/
  MemCopy ( ChData -> SubDeviceID , DeviceID , sizeof ( DeviceIDType ) ) ;
  ChData -> CurrentState = false ;
  if ( * DeviceID )
    ChData -> ValidDescr = true ;
  else
    ChData -> ValidDescr = false ;
} /* InitSingleChannel */


static void InitChannelsData ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    InitSingleChannel ( & ChannalsData [ Loop ] ,  EEBlock . SubDeviceID [ Loop ] ) ;
  } /* for */ ;
} /* InitChannelsData */


static uint8_t CompareFlash ( uint8_t * Text ,
                              uint16_t  Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * Text ++ ;
    Ch2 = pgm_read_byte_near ( Pattern ++ ) ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( Ch1 == 0 )
      return ( true ) ;
  } /* for */ ;
} /* CompareFlash */


static uint8_t CompareDeviceID ( uint8_t * DeviceName ,
                                 uint8_t * Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * DeviceName ++ ;
    Ch2 = * Pattern ++ ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( ( Ch1 == null ) || ( Ch2 == null ) )
      return ( true ) ;
  } /* for */ ;
} /* CompareDeviceID */


static uint8_t SuperiorDeviceDestination ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  return ( CompareDeviceID ( SerialInstance -> DestinationDeviceName , EEBlock . SuperiorDeviceID ) ) ;
} /* SuperiorDeviceDestination */


static uint8_t MeDestination ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  return ( CompareDeviceID ( SerialInstance -> DestinationDeviceName , EEBlock . MyDeviceID ) ) ;
} /* MeDestination */


static uint8_t GeneralDestination ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  return ( CompareFlash ( SerialInstance -> DestinationDeviceName , ( uint16_t ) GeneralPoolDeviceName ) ) ;
} /* GeneralDestination */


static void SendReplyForCommand ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SerialInstance -> SendSerial ( 'R' ) ;
  SendTxt ( SerialInstance , SerialInstance -> CommandText ) ;
  SerialInstance -> SendSerial ( ParamSeparator ) ;
} /* SendReplyForCommand */


static void SendReplyDevicesID ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SendTxt ( SerialInstance , SerialInstance -> SourceDeviceName ) ;
  SerialInstance -> SendSerial ( PartSeparator ) ;
  SendTxt ( SerialInstance , SerialInstance -> DestinationDeviceName ) ;
  SerialInstance -> SendSerial ( PartSeparator ) ;
} /* SendReplyDevicesID */


static void SendCommandReplyOK ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SerialInstance ) ;
  SendReplyForCommand ( SerialInstance ) ;
  SendFlashTxt ( SerialInstance , ( uint16_t ) OK ) ;
} /* SendCommandReplyOK */


static void SendCommandReplyFail ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SerialInstance ) ;
  SendReplyForCommand ( SerialInstance ) ;
  SendFlashTxt ( SerialInstance , ( uint16_t ) Fail ) ;
} /* SendCommandReplyFail */


static void ConnectCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = EEBlock . Password [ Loop ] ;
    if ( SerialInstance -> ParamText [ Loop ] != Ch )
    {
      SendCommandReplyFail ( SerialInstance ) ;
      return ;
    } /* if */ ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  ConnectContext = true ;
  SendCommandReplyOK ( SerialInstance ) ;
  LEDOn ( ConnectLED ) ;
} /* ConnectCommandExecute */


static void DisconnectCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  SendCommandReplyOK ( SerialInstance ) ;
  LEDOff ( ConnectLED ) ;
} /* DisconnectCommandExecute */


static uint8_t ExecuteAssociatedCommand ( PtrSerialInstanceRecType SerialInstance ,
                                          uint8_t *                String ,
                                          uint8_t                  AssociationTableSize ,
                                          uint16_t                 AssociationTable )
{
  uint8_t Loop ;
  CommandNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CommandService = ( CommandExecuteServiceProcT ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    AssociationElement . CommandName = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . CommandName ) )
    {
      AssociationElement . CommandService ( SerialInstance ) ;
      return ( true ) ;
    } /* if */ ;
  } /* for */ ;
  return ( false ) ;
} /* ExecuteAssociatedCommand */


static void HelloCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SerialInstance ) ;
  SendReplyForCommand ( SerialInstance ) ;
  SendFlashTxt ( SerialInstance , ( uint16_t ) DeviceName ) ;
} /* HelloCommandExecute */


static void GetVersionCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SerialInstance ) ;
  SendReplyForCommand ( SerialInstance ) ;
  SendFlashTxt ( SerialInstance , ( uint16_t ) HardwareSoftwareVersion ) ;
} /* GetVersionCommandExecute */


static void GetIDCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SerialInstance ) ;
  SendReplyForCommand ( SerialInstance ) ;
  SendTxt ( SerialInstance , EEBlock . MyDeviceID ) ;
  SerialInstance -> SendSerial ( cr ) ;
} /* GetIDCommandExecute */


static void ReinitCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlockOperation ( ) ;
  InitChannelsData ( ) ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* ReinitCommandExecute */


static void ResetEECommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  SetDefaultEEBlock ( ) ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = CRCValue ;
  InitChannelsData ( ) ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* ResetEECommandExecute */


static void SaveEECommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  LEDOn ( EEWriteLED ) ;
  EEBlock . CRCValue = 0 ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = CRCValue ;
  WriteEEBlock ( ) ;
  SendCommandReplyOK ( SerialInstance ) ;
  LEDOff ( EEWriteLED ) ;
} /* SaveEECommandExecute */


static void Assign ( uint8_t *                SubDeviceID ,
                     PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Loop ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  ClearBlock ( SubDeviceID , sizeof ( DeviceIDType ) ) ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SerialInstance -> ParamText [ Loop ] ;
    if ( ! Ch )
      break ;
    * SubDeviceID ++ = Ch ;
  } /* for */ ;
} /* Assign */


static void AddSlaveIDCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialInstance -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SerialInstance ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( ! EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      Assign ( EEBlock . SubDeviceID [ Loop ] , SerialInstance ) ;
      InitChannelsData ( ) ;
      SendCommandReplyOK ( SerialInstance ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SerialInstance ) ;
} /* AddSlaveIDCommandExecute */


static void DelSlaveIDCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialInstance -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SerialInstance ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( EEBlock . SubDeviceID [ Loop ] , SerialInstance -> ParamText ) )
    {
      ClearBlock ( EEBlock . SubDeviceID [ Loop ] , sizeof ( DeviceIDType ) ) ;
      InitChannelsData ( ) ;
      SendCommandReplyOK ( SerialInstance ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SerialInstance ) ;
} /* DelSlaveIDCommandExecute */


static void ListSlaveIDCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Loop ;
  uint8_t AnyDataPresent ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  AnyDataPresent = false ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( ChannalsData [ Loop ] . ValidDescr )
    {
      AnyDataPresent = true ;
      SendReplyDevicesID ( SerialInstance ) ;
      SendReplyForCommand ( SerialInstance ) ;
      SendTxt ( SerialInstance , ChannalsData [ Loop ] . SubDeviceID ) ;
      SerialInstance -> SendSerial ( cr ) ;
    } /* if */ ;
  } /* for */ ;
  if ( ! AnyDataPresent )
  {
    SendReplyDevicesID ( SerialInstance ) ;
    SendReplyForCommand ( SerialInstance ) ;
    SendFlashTxt ( SerialInstance , ( uint16_t ) EmptyData ) ;
  } /* if */ ;
} /* ListSlaveIDCommandExecute */


static void SetDeviceNameCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialInstance -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SerialInstance ) ;
    return ;
  } /* if */ ;
  Assign ( EEBlock . MyDeviceID , SerialInstance ) ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* SetDeviceNameCommandExecute */


static void AddSuperiorChnCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialInstance -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SerialInstance ) ;
    return ;
  } /* if */ ;
  Assign ( EEBlock . SuperiorDeviceID , SerialInstance ) ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* AddSuperiorChnCommandExecute */


static void DelSuperiorChnCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialInstance -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SerialInstance ) ;
    return ;
  } /* if */ ;
  if ( CompareDeviceID ( EEBlock . SuperiorDeviceID , SerialInstance -> ParamText ) )
  {
    ClearBlock ( EEBlock . SuperiorDeviceID , sizeof ( DeviceIDType ) ) ;
    SendCommandReplyOK ( SerialInstance ) ;
    return ;
  } /* if */ ;
  SendCommandReplyFail ( SerialInstance ) ;
} /* DelSuperiorChnCommandExecute */


static void ListSuperiorChnCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( EEBlock . SuperiorDeviceID [ 0 ] )
  {
    SendReplyDevicesID ( SerialInstance ) ;
    SendReplyForCommand ( SerialInstance ) ;
    SendTxt ( SerialInstance , EEBlock . SuperiorDeviceID ) ;
    SerialInstance -> SendSerial ( cr ) ;
  } /* if ... */ 
  else
  {
    SendReplyDevicesID ( SerialInstance ) ;
    SendReplyForCommand ( SerialInstance ) ;
    SendFlashTxt ( SerialInstance , ( uint16_t ) EmptyData ) ;
  } /* if ... else */ ;
} /* ListSuperiorChnCommandExecute */


static void NewPasswordCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = SerialInstance -> ParamText [ Loop ] ;
    EEBlock . Password [ Loop ] = Ch ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* NewPasswordCommandExecute */


static void EchoOnCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SerialRS232Instance . EchoSerial = USART232SendSerial ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* EchoOnCommandExecute */


static void EchoOffCommandExecute ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  SerialRS232Instance . EchoSerial = NoEcho ;
  SendCommandReplyOK ( SerialInstance ) ;
} /* EchoOffCommandExecute */


static void RS232ExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( & SerialRS232Instance ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( & SerialRS232Instance ) ;
    if ( FrameToMe )
      StrCopy ( SerialRS232Instance . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
    if ( ! ExecuteAssociatedCommand ( & SerialRS232Instance , SerialRS232Instance . CommandText ,
                                      SvCommandNameAssociationTabeSize ,
                                      ( uint16_t ) SvCommandNameAssociationTabe ) )
    {
    } /* if */ ;
  } /* if ... */
  else
  {
    SendTxt ( & SerialRS485Instance , SerialRS232Instance . DestinationDeviceName ) ;
    SerialRS485Instance . SendSerial ( PartSeparator ) ;
    SendTxt ( & SerialRS485Instance , SerialRS232Instance . SourceDeviceName ) ;
    SerialRS485Instance . SendSerial ( PartSeparator ) ;
    SendTxt ( & SerialRS485Instance , SerialRS232Instance . CommandText ) ;
    if ( SerialRS232Instance . ParamText [ 0 ] )
    {
      SerialRS485Instance . SendSerial ( ParamSeparator ) ;
      SendTxt ( & SerialRS485Instance , SerialRS232Instance . ParamText ) ;
    } /* if */ ;
    SerialRS485Instance . SendSerial ( cr ) ;
  } /* if ... else */ ;
} /* RS232ExecuteCommand */


static void RS485ExecuteCommand ( void )
{
  PtrSerialInstanceRecType SerialInstance ;
  /*----------------------------------------------------------------*/
  if ( SuperiorDeviceDestination ( & SerialRS485Instance ) )
  {
    SerialInstance = & SerialRS232Instance ;
    SendTxt ( SerialInstance , SerialRS485Instance . DestinationDeviceName ) ;
    SerialInstance -> SendSerial ( PartSeparator ) ;
    SendTxt ( SerialInstance , SerialRS485Instance . SourceDeviceName ) ;
    SerialInstance -> SendSerial ( PartSeparator ) ;
    SendTxt ( SerialInstance , SerialRS485Instance . CommandText ) ;
    if ( SerialRS485Instance . ParamText [ 0 ] )
    {
      SerialInstance -> SendSerial ( ParamSeparator ) ;
      SendTxt ( SerialInstance , SerialRS485Instance . ParamText ) ;
    } /* if */ ;
    SerialInstance -> SendSerial ( cr ) ;
  } /* if */ ;
} /* RS485ExecuteCommand */


static void ResetOperation ( PtrSerialInstanceRecType SerialInstance ) ;

static void DestinationIDPutCh ( PtrSerialInstanceRecType SerialInstance ,
                                 uint8_t                  Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> DestDeviceNameIndex < DeviceIDSize )
  {
    SerialInstance -> DestinationDeviceName [ SerialInstance -> DestDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* DestinationIDPutCh */


static void SourceIDPutCh ( PtrSerialInstanceRecType SerialInstance ,
                            uint8_t                  Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> SrcDeviceNameIndex < DeviceIDSize )
  {
    SerialInstance -> SourceDeviceName [ SerialInstance -> SrcDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* SourceIDPutCh */


static void CommandIDPutCh ( PtrSerialInstanceRecType SerialInstance ,
                             uint8_t                  Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> CommandTextIndex < CommandBufferSize )
  {
    SerialInstance -> CommandText [ SerialInstance -> CommandTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* CommandIDPutCh */


static void ParamPutCh ( PtrSerialInstanceRecType SerialInstance ,
                         uint8_t                  Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> ParamTextIndex < ParamBufferSize )
  {
    SerialInstance -> ParamText [ SerialInstance -> ParamTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* ParamPutCh */


static void SelectParam ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> SerialDataPresent ( ) )
  {
    Ch = SerialInstance -> GetSerial ( ) ;
    SerialInstance -> EchoSerial ( Ch ) ;
    if ( Ch == cr )
    {
      ParamPutCh ( SerialInstance , null ) ;
      SerialInstance -> ExecuteCommand ( ) ;
      ResetOperation ( SerialInstance ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( SerialInstance ) ;
      return ;
    } /* if */ ;
    ParamPutCh ( SerialInstance , Ch ) ;
  } /* if */ ;
} /* SelectParam */


static void SelectCommand ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> SerialDataPresent ( ) )
  {
    Ch = SerialInstance -> GetSerial ( ) ;
    SerialInstance -> EchoSerial ( Ch ) ;
    if ( Ch == cr )
    {
      CommandIDPutCh ( SerialInstance , null ) ;
      SerialInstance -> ExecuteCommand ( ) ;
      ResetOperation ( SerialInstance ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( SerialInstance ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
    {
      CommandIDPutCh ( SerialInstance , null ) ;
      SerialInstance -> SerialService = SelectParam ;
      return ;
    } /* if */;
    CommandIDPutCh ( SerialInstance , Ch ) ;
  } /* if */ ;
} /* SelectCommand */


static void SelectSourceService ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> SerialDataPresent ( ) )
  {
    Ch = SerialInstance -> GetSerial ( ) ;
    SerialInstance -> EchoSerial ( Ch ) ;
    if ( Ch < Space )
    {
      ResetOperation ( SerialInstance ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      SourceIDPutCh ( SerialInstance , null ) ;
      SerialInstance -> SerialService = SelectCommand ;
      return ;
    } /* if */;
    SourceIDPutCh ( SerialInstance , Ch ) ;
  } /* if */ ;
} /* SelectSourceService */


static void SelectDestOperation ( PtrSerialInstanceRecType SerialInstance )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialInstance -> SerialDataPresent ( ) )
  {
    Ch = SerialInstance -> GetSerial ( ) ;
    SerialInstance -> EchoSerial ( Ch ) ;
    if ( Ch < Space )
    {
      ResetOperation ( SerialInstance ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      DestinationIDPutCh ( SerialInstance , null ) ;
      SerialInstance -> SerialService = SelectSourceService ;
      return ;
    } /* if */;
    DestinationIDPutCh ( SerialInstance , Ch ) ;
  } /* if */ ;
} /* SelectDestOperation */


static uint8_t SelectStartDevIndex ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      return Loop ;
    } /* if */ ;
  } /* for */ ;
  return 0xFF ;
} /* SelectStartDevIndex */


static uint8_t SelectNextDevice ( uint8_t StartIndex )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( StartIndex == 0xFF)
    return 0xFF ;
  StartIndex ++ ;
  if ( StartIndex >= ChannelsArraySize )
    StartIndex = 0 ;
  for ( Loop = StartIndex ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      return Loop ;
    } /* if */ ;
  } /* for */ ;
  return SelectStartDevIndex ( ) ;
} /* SelectNextDevice */


static uint8_t SendReplyReq ( uint8_t DevIndex )
{
  /*----------------------------------------------------------------*/
  if ( DevIndex != 0xFF )
  {
    SendTxt ( & SerialRS485Instance , EEBlock . SubDeviceID [ DevIndex ] ) ;
    SerialRS485Instance . SendSerial ( PartSeparator ) ;
    SendTxt ( & SerialRS485Instance , EEBlock . MyDeviceID ) ;
    SerialRS485Instance . SendSerial ( PartSeparator ) ;
    SendFlashTxt ( & SerialRS485Instance , ( uint16_t ) ReplyReqCommand ) ;
    SerialRS485Instance . SendSerial ( cr ) ;
    return true ;
  } /* if */ ;
  return false ;
} /* SendReplyReq */


static void SendQueueElement ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( USART485 . TXDCycFr != USART485 . TXDCycTo )
  {
    cli ( ) ;
    Ch = USART485 . TXDBuffer [ USART485 . TXDCycFr ] ;
    USART485 . TXDCycFr ++ ;
    if ( USART485 . TXDCycFr >= TXD485BuffSize )
      USART485 . TXDCycFr = 0 ;
    USART485 . TXDState = BusyState ;
    sei ( ) ;
    USART485TransmitMode ( ) ;
    StartUSART485RxTxLED ( ) ;
    UDR1 = Ch ;
  } /* if */ ;
} /* SendQueueElement */


static void ResetOperation ( PtrSerialInstanceRecType SerialInstance )
{
  /*----------------------------------------------------------------*/
  ClearBlock ( SerialInstance -> CommandText , sizeof ( CommandBufferType ) ) ;
  ClearBlock ( SerialInstance -> ParamText , sizeof ( ParamBufferType ) ) ;
  ClearBlock ( SerialInstance -> DestinationDeviceName , sizeof ( DeviceIDType ) ) ;
  ClearBlock ( SerialInstance -> SourceDeviceName , sizeof ( DeviceIDType ) ) ;
  SerialInstance -> CommandTextIndex = 0 ;
  SerialInstance -> ParamTextIndex = 0 ;
  SerialInstance -> DestDeviceNameIndex = 0 ;
  SerialInstance -> SrcDeviceNameIndex = 0 ;
  SerialInstance -> SerialService = SelectDestOperation ;
} /* ResetOperation */


static void SoftwareInit ( void )
{
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  CurrentTimeStatusState = 0 ;
  USART0Reset ( ) ;
  USART1Reset ( ) ;
  SerialRS232Instance . SendSerial = USART232SendSerial ;
  SerialRS232Instance . EchoSerial = NoEcho ;
  SerialRS232Instance . SerialDataPresent = USART232DataPresent ;
  SerialRS232Instance . GetSerial = USART232GetSerial ;
  SerialRS232Instance . ExecuteCommand = RS232ExecuteCommand ;
  SerialRS485Instance . SendSerial = USART485SendSerial ;
  SerialRS485Instance . EchoSerial = NoEcho ;
  SerialRS485Instance . SerialDataPresent = USART485DataPresent ;
  SerialRS485Instance . GetSerial = USART485GetSerial ;
  SerialRS485Instance . ExecuteCommand = RS485ExecuteCommand ;
  TimerCounter = 0 ;
  ResetOperation ( & SerialRS232Instance ) ;
  ResetOperation ( & SerialRS485Instance ) ;
  Events = 0 ;
  SubDevTimerCounter = 0 ;
  QueueTimerCounter = 0 ;
} /* SoftwareInit */


int main ( void )
{
  /*----------------------------------------------------------------*/
  HardwareInit ( Speed38400bps , Speed9600bps ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  EEBlockOperation ( ) ;
  InitChannelsData ( ) ;
  USART485ReceiveMode ( ) ;
  wdt_enable ( WDTO_1S ) ;
  SubDeviceIndex = SelectStartDevIndex ( ) ;
  sei ( ) ;
  for ( ; ; )
  {
    wdt_reset ( ) ;
    if ( Events & TimeEvent )
    {
      Events &= ~ TimeEvent ;
      if ( CurrentTimeStatusState )
        LEDOn ( StatusLED ) ;
      else
        LEDOff ( StatusLED ) ;
      CurrentTimeStatusState = 1 - CurrentTimeStatusState ;
    } /* if */ ;
    if ( Events & SubDeviceEvent )
    {
      Events &= ~ SubDeviceEvent ;
      if ( SendReplyReq ( SubDeviceIndex ) )
        SubDeviceIndex = SelectNextDevice ( SubDeviceIndex ) ;
    } /* if */ ;
    if ( Events & QueueEvent )
    {
      Events &= ~ QueueEvent ;
      SendQueueElement ( ) ;
    } /* if */ ;
    if ( SerialRS232Instance . SerialService )
      SerialRS232Instance . SerialService ( & SerialRS232Instance ) ;
    if ( SerialRS485Instance . SerialService )
      SerialRS485Instance . SerialService ( & SerialRS485Instance ) ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


